% ----------------------------------------------------------------------
%
% Tcl utilities for encoding and decoding attributes.
%
% Author: Drew van Camp (drew@@cs.toronto.edu)
%
% Copyright (c) 1995-1996 The University of Toronto.
%
% See the file "copyright" for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL WARRANTIES.
%
% ----------------------------------------------------------------------

\documentclass{article}
\usepackage{moretext,amstex,alltt}
\usepackage{delve}

\newcommand{\acdc}{\textbf{Acdc}}

\begin{document}

\rcsInfo $Id: acdc.w,v 1.9.2.8 1996/06/14 16:06:46 drew Exp $

\title{Acdc: Attribute Encoding/Decoding Procedures}

\author{Drew van Camp (drew@@@@cs.toronto.edu)\\[1ex]
	Department of Computer Science\\
	University of Toronto\\
	6 Kings College Road\\
	Toronto ON, Canada, M5S 1A4}

\vfil
\maketitle
\vfil
\copyrightNotice{1996}
\vfil
\clearpage
\tableofcontents
\clearpage

\section{Introduction}
The \acdc\ module contains procedures for managing the encoding and
decoding of attributes. Encoding is a complicated and potentially time
consuming area. The internals of the module may be complicated to
enhance speed; however, it is hoped that the interface is simple to
use.

\section{The \texttt{acdc\_new} command} \label{sec:interface:acdcNew}
\subsection{Description}
\usage{acdc\_new}{type id ?alias ...?}
Adds a new attribute to the encoding list. You must supply a
\texttt{type} for the attribute and a unique \texttt{id} for it.
As well, you may specify any number of aliases for the
\texttt{id}. Id's and aliases can be used interchangably as the
\texttt{attr} argument to all other commands in the \acdc\ module. 

If the attribute \texttt{id} is already in use, it will be silently
replaced. Similarly, if any \texttt{alias} is already in use, it will
be replaced.

Valid \texttt{types} for the attribute are:
\begin{description}
\item[\texttt{angular}] - attributes that have a circular nature, in that
	only the variable modulus some length is thought to be
	important.
\item[\texttt{binary}] - attributes that take on only two possible
	values.
\item[\texttt{nominal}] - nominal attributes are significant only in
    	that they may take on a finite number of distinct values.
\item[\texttt{ordinal}] - oridinal attributes may take on a finite
	number of distinct values and they have a defined ordering,
	which must be specified.
\item[\texttt{integer}] - integer attributes take on only integer values.
\item[\texttt{real}] - real attributes take on any real value.
\end{description}

\subsection{Implementation}
The procedure takes as arguments: \texttt{type}, the type of the
attribute (as described above); \texttt{id}, a unique id for the
attribute to be encoded; and \texttt{aliases}, a list of symbolic
aliases that can be used interchangeably with the attribute id.  The
general structure of the procedure is as follows:
 
@o acdc.tcl @{
proc acdc_new { type id args } {
    @<get access to state information@>
    @<delete the attribute if it already exists@>
    @<initialize attribute parameters@>
    @<initialize encoding method@>
    @<initialize encoding options@>
}
@}

All state information is stored in the global array
\texttt{acdc\_priv}. This array, as its name implies, is a private
array and should \textit{never} be directly accessed by external
procedures.

Since there is only one array, only one set of attributes can be
encoded by an application at any one time. This greatly simplifies the
interface. For each attribute, there are several elements stored in
the state array. All elements that correspond to an attribute have
keys of the form \texttt{id,key} where \texttt{id} is the id of the
attribute, and \texttt{key} is the sub-key of the field, for example
\texttt{type}. As well, there are entries that map aliases to
ids. These entries are of the form \texttt{acdc\_id,alias}, where
\texttt{acdc\_id} is a literal string and \texttt{alias} is the alias
string. The value stored in this variable is the id for the attribute.

@d get access to state information @{
global acdc_priv ;
@}

If the attribute already exists, we silently delete it and replace it
with the new one. To do this we check if there is an id stored in the
global state array for the new id.

@d delete the attribute if it already exists @{
if [info exists acdc_priv(acdc_id,$id)] {
    acdc_delete $id ;
}
@}

Initializing the attribute parameters entails storing the arguments
passed to the procedure in the global state array \texttt{acdc\_priv}:
the type, the id, and all of the aliases.

@d initialize attribute parameters @{
set acdc_priv($id,type)		$type ;
set acdc_priv(acdc_id,$id)	$id ;
foreach alias $args {
    set acdc_priv(acdc_id,$alias)	$id ;
}
@}

Next we initialize the encoding method for the attribute. To do this
we use the private procedure \texttt{acdcDefault}, which returns the
default encoding for an attribute of \texttt{type}. We also set the
active method to the default method.

@d initialize encoding method @{
set acdc_priv($id,defaultMethod) [acdcDefault $type] ;
set acdc_priv($id,method)	 $acdc_priv($id,defaultMethod) ;
@}

Finally, we must initialize all of the options for the encoding
methods. Their default values are all empty strings (which means they
have not been specified).

@d initialize encoding options @{
foreach option [list absDev center mean median passive range variance unit] {
    set acdc_priv($id,$option) 	{} ;
}
set acdc_priv($id,scale) 	sqrt ;
@}

\subsubsection{The \texttt{acdcDefault} command}
\usage{acdcDefault}{type}
This procedure returns the default encoding method for a \texttt{type}
of attribute. This is different from the default encoding method for
an attribute. In fact, it is the initial default encoding method for
an attribute. The procedure is implemented as a simple switch:

@o acdc.tcl @{
proc acdcDefault { type } {
    switch -- $type {
        angular         { return "rectan" }
        binary          { return "-1/+1" }
        nominal         { return "1-of-n" }
        ordinal         { return "therm" }
        integer         { return "nm-abs" }
        real            { return "nm-abs" }
        default         { error "can't encode attribute type \"$type\"" }
    }
}
@}

\section{The \texttt{acdc\_delete} command}
\subsection{Description}
\usage{acdc\_delete}{attr}
Deletes the attribute \texttt{attr} from the encoding list.  If no
such attibute exists, it returns without changing the list. The
\texttt{attr} argument may be either an attribute's id, or any of its
aliases.

\subsection{Implementation}
This procedure takes one argument, \texttt{attr} the attribute to
delete. If the attribute doesn't exist, then the routine returns
without doing anything.  The overall structure of the procedure is:

@o acdc.tcl @{
proc acdc_delete { attr } {
    @<get access to state information@>
    @<if the attribute doesn't exist, return@>
    @<remove all state information about this attribute@>
}
@}

The state information is retrieved as in \texttt{acdc\_new}, then we
see if information about the attribute is stored in the state
array. If not, we just return.

@d if the attribute doesn't exist, return @{
if ![info exists acdc_priv(acdc_id,$attr)] {
    return ;
}
@}

Removing all state information about the attribute is done by
unsetting all variables in \texttt{acdc\_priv} relating to it and
deleting any compiled encodings (see Section~\ref{sec:acdcCompile}).
All of the variables have array keys \texttt{\$id,key} where
\texttt{\$id} is the id for the attribute and \texttt{key} can be any
value.  We do simple \texttt{glob} pattern matching to find all these
keys.  We also have to remove any aliases that map to the id.  To do
that we have to iterate over all keys that have the form
\texttt{acdc\_id,alias} and remove any that map to
\texttt{\$id}.

@d remove all state information about this attribute @{
set id $acdc_priv(acdc_id,$attr) ;
if [info exists acdc_prive($id,handle)] {
    d_attr delete $acdc_prive($id,handle)
}
foreach key [array names acdc_priv] {
    if [regexp "^$id," $key] {
	unset acdc_priv($key) ;
    }
}
foreach key [array names acdc_priv] {
    if { [regexp "^acdc_id," $key] && [string match $acdc_priv($key) $id] } {
	unset acdc_priv($key) ;
    }
}
@}

\section{acdc\_method}
\subsection{Description}
\usage{acdc\_method}{attr ?switches? ?method?}
Returns the encoding method used for the attribute \texttt{attr}.  If
the optional \texttt{method} is specified, the appropriate method for
the attribute is set to \texttt{method} before returning the value. If
the attribute can't be encoded using \texttt{method}, an error is
generated and the method is unchanged.

If the initial arguments to \texttt{acdc\_method} start with - then
they are treated as switches.  The following switches are currently
supported:

\begin{description}
\item[\texttt{-default}] If this switch is specified, the procedure
	operates on the default encoding method (which may be
	different than the \textit{active} encoding method (the one
	actually being used).
\item[\texttt{--}] Marks the end of switches.  The argument following
	this one will be treated as \texttt{method} even if it starts
	with a -.
\end{description}

Valid methods and their options are:
\begin{description}
\item[\texttt{copy}] Does not encode an attribute, but returns the
	original value. This encoding is valid for all attribute types
	and takes no options
\item[\texttt{ignore}] Encodes an attribute as an empty string. This
	is valid for all attribute types. Since it is a many to one
	encoding, it cannot be decoded. 
\item[\texttt{-1/+1}] Uses symmetric binary encoding. This is only valid
	for attributes of type \texttt{binary}. This encoding uses the
	\texttt{-range} option which should contain a list of the two
	valid states of the attribute. The first state will encode as
	-1, the second as +1.
\item[\texttt{0/1}] Encodes \texttt{binary} attributes as an
	active/passive value. This encoding uses both the
	\texttt{-range} option to get the valid states for the
	attribute, and the \texttt{-passive} option to determine which
	state maps to zero.
\item[\texttt{0-up}] Encodes an attribute as an integer, ranging from
	zero and up. This encoding is valid for \texttt{norminal} and
	\texttt{ordinal} variables. It uses the \texttt{-range} option
	to determine the ordering of the mapping. The range should
	contain a list of all valid states of the attribute, in the
	order they should be mapped to integers.
\item[\texttt{1-up}] Encodes an attribute as an integer, ranging from
	one and up. This is the same as the \texttt{0-up} encoding,
	offset by one.
\item[\texttt{1-of-n}] Uses \texttt{n} binary active/passive values,
	turning on a single value for each pattern. This encoding is
	valid for \texttt{nominal} and \texttt{ordinal}
	attributes. The number \texttt{n} is determined from the
	number of states in the \texttt{-range} option. If the
	optional \texttt{-passive} is given, the
	corresponding value is omitted (i.e. there are \texttt{n-1}
	values, and the passive value encodes to all zeroes).
\item[\texttt{therm}] Encodes using a thermometer code with
        \texttt{n-1} binary symmetric inputs. For the lowest value of
	the attribute only the no units are on, for the next higher
	value the first unit is on, etc.  This encoding is valid for
	\texttt{nominal} and \texttt{ordinal} attributes. The number
	\texttt{n} and the order of the states is determined from the
	\texttt{-range} option.  The \texttt{-scale} option may be
	supplied with one of the values \texttt{none}, \texttt{sqrt},
	\texttt{linear}.  If \texttt{-scale} is \texttt{sqrt}, the
	values in the encoding are divided by $\sqrt{n-1}$.  If it is
	\texttt{linear}, they are divided by $n-1$, and if it is
	\texttt{none}, they are not divided.
\item[\texttt{nm-abs}] Normalise a value to zero median and unit
	absolute deviation from the median. This encoding is valid for
	\texttt{real} and \texttt{integer} attributes. It requires you
	to specify a \texttt{-median} and \texttt{-absDev}.
	Optionally, you may specify a \texttt{-centre} value to
	override zero as the median.
\item[\texttt{nm-sqr}] Normalise a value to zero mean and unit
	variance. This encoding is valid for \texttt{real} and
	\texttt{integer} attributes. It requires you to specify a
	\texttt{-mean} and \texttt{-variance}.  Optionally, you may
	specify a \texttt{-centre} value to override zero as the
	median.
\item[\texttt{rectan}] Encodes \texttt{angular} attributes using two
	values, $\sin(2\pi x/{\rm unit})$ and $\cos(2\pi x/{\rm
	unit})$, where $x$ is the value of the attribute and ${\rm
	unit}$ is a value specified using \texttt{-unit}. This is a
	many to one encoding and, as such, is not decodable.
\end{description}

\subsection{Implementation}
The procedure has one required argument, \texttt{attr}, the attribute
of interest, and one optional argument: a \texttt{method} to set the
encoding to.  As well, there are optional switches for the command.
Both the optional arguments and switches are passed in the
\texttt{args} parameter:

@o acdc.tcl @{
proc acdc_method { attr args } {
    @<get access to state information@>
    @<get the attribute id@>
    @<remove all switches from the args option@>
    @<check that we have the right number of remaining args@>
    @<see if we're using the active or default method@>
    @<if a method was specified, change it@>
    @<return the method@>
}
@}

As usual, we need access to the \texttt{acdc\_priv} variable. Once we
have it, check if the attribute exists, and generate an error if not.
Many routines generate an error if the attribute passed to them
doesn't exist, so it's worth keeping this bit of code separate:

@o acdc.tcl @{
proc acdcGetId { attr } {
    if ![info exists acdc_priv(acdc_id,$attr)] {
	error "attribute \"$attr\" hasn't been registered for encoding" ;
    } 
    return $acdc_priv(acdc_id,$attr) ;
}
@}

@d get the attribute id @{
if ![info exists acdc_priv(acdc_id,$attr)] {
    error "attribute \"$attr\" hasn't been registered for encoding" ;
} 
set id $acdc_priv(acdc_id,$attr) ;
@}

Next, we iterate over the \texttt{args} parameter removing all
switches. Switches must come before all other arguments. The switches are
terminated by the argument \texttt{--}, or the first argument that
doesn't with a \texttt{-}. The only switch that we recognize is
\texttt{-default}. Any other one will generate an error.

@d remove all switches from the args option @{
while { ![lempty $args] } {
    set switch [lpop args] ;
    switch -glob -- $switch {
	--		{ break ; }
	-default	{ set default 1 ; }
	-*		{ error "unknown switch \"$switch\"" }
	default		{ lpush args $switch ; break ; }
    }
}
@}

After the switches have been removed from \texttt{args}, it should be
either empty, or have one argument: the name of the desired method. 

@d check that we have the right number of remaining args @{
if { [llength $args] > 1 } {
    error "wrong \# args: should be \"acdc_method attr ?switches? ?method?\"" ;
}
@}

If the \texttt{default} variable has been set (from the optional
switches), then we are interested in the default encoding method. If not,
we are interested in the active method. We just save the key into the
\texttt{acdc\_priv} array of whichever one we'll be using.

@d see if we're using the active or default method @{
if [info exists default] {
    set key defaultMethod ;
} else {
    set key method ;
}
@}

If there is one argument left in args, a method was specified on the
command line: we save it to a local variable \texttt{method}.  As
well, we save the type of the attribute to the variable \texttt{type}.
Once we have both of these, we check if the method is consistent with
the type.  The private procedure \texttt{acdcAllowed} (see
Section~\ref{sec:acdcAllowed}) returns a list of all allowed encodings
for an attribute type.  We search the list of allowed methods for the
attribute's type, seeing if the given method is in it. If not, we
generate an error.  If it is a valid type, we save it in the state
information.  Now we have to be careful.  Since we've changed the
encoding method, we have to throw away the compiled form of the old
method (if it exists) by deleting the encoding handle and unsetting
the variable containing it.

@d if a method was specified, change it @{
if { [llength $args] == 1 } {
    set method	[lpop args] ;
    set type	$acdc_priv($id,type) ;
    if { [lsearch [acdcAllowed $type] $method] < 0 } {
	error "can't use \"$method\" encoding for \"$type\" attribute $attr" ;
    }
    set acdc_priv($id,$key)	$method ;
    if [info exists acdc_priv($id,handle)] {
	acdc_delete $acdc_priv($id,handle) ;
	unset acdc_priv($id,handle) ;
    }
}
@}

Now we're at the end of the procedure.  The final line of the
procedure returns the appropriate method for the attribute (which we
may have just changed).

@d return the method @{
return $acdc_priv($id,$key) ;
@}

\subsubsection{The \texttt{acdcAllowed}}
\label{sec:acdcAllowed}
\usage{acdcAllowed}{type}
This \textit{private} procedure returns a list of all encoding methods
allowed for attributes of \texttt{type}. As in \texttt{acdcDefault},
it is implemented as a simple switch.

@o acdc.tcl @{
proc acdcAllowed { type } {
    switch -- $type {
	angular     { return "ignore copy rectan" }
	binary      { return "ignore copy -1/+1 0/1" }
	nominal     { return "ignore copy 1-of-n 0-up 1-up therm" }
	ordinal     { return "ignore copy 1-of-n 0-up 1-up therm" }
	integer     { return "ignore copy nm-abs nm-sqr" }
	real        { return "ignore copy nm-abs nm-sqr" }
	default     { error "can't encode attribute type \"$type\"" }
    }
}
@}

\section{The \texttt{acdc\_options} command}
\subsection{Description}
\usage{acdc\_options}{attr option ?value option value ...?}
Some encoding methods have optional (or required) parameters. These
parameters can be queried and manipulated using \texttt{acdc\_options}.
If a single option is given to this procedure, the current value of
that option for the attribute \texttt{attr} is returned. If one or
more \texttt{option} \texttt{value} pairs is given as arguments, then
the options are set to the values, and an empty string is
returned. Valid options are described under the different methods
used by the \texttt{acdc\_method} command. 

\subsection{Implementation}
This procedure has two required arguments: the attribute
\texttt{attr}, and one option. After initialization, the procedure has
two main branches. The first branch occurs when a single switch is
given on the command line. In this case we return the value of the
option corresponding to the switch. The second branch occurs when one
or more option-value pairs are given on the command line. In this case
we iterate over all of the switches, setting the appropriate
values. Once we've done that we have to delete the compiled encoding
(it may not be valid after the options have changed). Finally, we
return an empty string.

@o acdc.tcl @{
proc acdc_options { attr args } {
    @<get access to state information@>
    @<get the attribute id@>
    @<if there are no arguments, generate an error@>

    if { [llength $args] == 1 } {
	@<return the value of the appropriate option@>
    } else {
	@<iterate over all arguments setting the appropriate options@>
	if [info exists acdc_priv($id,handle)] {
	    acdc_delete $acdc_priv($id,handle) ;
	    unset acdc_priv($id,handle) ;
	}
	@<return an empty string@>
    }
}
@}

Beyond the normal initialization, we also check that at least one
argument was given after \texttt{attr}. If not, we generate an error
describing the usage.

@d if there are no arguments, generate an error @{
if { [llength $args] == 0 } {
    error "wrong \# args: should be \"acdc_options attr option ?value option value ...?\"" ;
}
@}

If only a single argument was given, do a switch on its value,
returning the state information associated with it, or generating an
error if we don't recognize the switch. We also allow the British
spelling of the \texttt{center} option (i.e. \texttt{centre}).  For
backwards compatability, we also support \texttt{stdDev}.

@d return the value of the appropriate option @{
set option [lpop args] ;
switch -glob -- $option {
    -absDev	{ return $acdc_priv($id,absDev) }
    -center	-
    -centre	{ return $acdc_priv($id,center) }
    -mean	{ return $acdc_priv($id,mean) }
    -median	{ return $acdc_priv($id,median) }
    -passive	{ return $acdc_priv($id,passive) }
    -range	{ return $acdc_priv($id,range) }
    -scale	{ return $acdc_priv($id,scale) }
    -unit	{ return $acdc_priv($id,unit) }
    -variance	{ return $acdc_priv($id,variance) }
    -stdDev	{
	if { [string compare $acdc_priv($id,variance) {}] == 0 } {
	    return {} ;
	} else {
	    return [expr { sqrt($acdc_priv($id,variance)) }] ;
	}
    }
    default	{ error "unknown option \"$option\"" }
}
@}

If multiple arguments were given, we have to iterate over all of them
\textit{setting} the appropriate values. We generate errors if a switch
is given with no corresponding value, or if we don't recognize a switch.

@d iterate over all arguments setting the appropriate options @{
while { ![lempty $args] } {
    set option [lpop args] ;
    if [lempty $args] {
	error "missing value for option \"$option\"" ;
    }
    set value  [lpop args] ;
    switch -glob -- $option {
	-absDev	  { set acdc_priv($id,absDev)	$value }
	-center	  -
	-centre	  { set acdc_priv($id,center)	$value }
	-mean	  { set acdc_priv($id,mean)	$value }
	-median	  { set acdc_priv($id,median)	$value }
	-passive  { set acdc_priv($id,passive)	$value }
	-range	  { set acdc_priv($id,range)	$value }
	-unit	  { set acdc_priv($id,unit)	$value }
	-scale	  {
	    if { [lsearch "none linear sqrt" $value] >= 0 } {
		set acdc_priv($id,scale)	$value ;
	    } else {
		error "bad scale option for \"$id\", \"$value\": should be none, linear, or sqrt" ;
	    }
	}
	-variance { set acdc_priv($id,variance)	$value }
	-stdDev	  {
	    if { [string compare $value {}] == 0 } {
		set acdc_priv($id,variance)	{} ;
	    } else {
		set acdc_priv($id,variance)	[expr { $value * $value }] ;
	    }
	}
	default	  { error "unknown option \"$option\"" }
    }
}
@}

In the second branch (multiple options) we have to be careful to
return an empty string:

@d return an empty string @{
    return ;
@}

\section{The \texttt{acdc\_encode} command}
\subsection{Description}
\usage{acdc\_encode}{attr value}
Returns the encoded representation of \texttt{value} using the
encoding method and options currently specified for the attribute
\texttt{attr}.

\subsection{Implementation}
Since the work involved in encoding a value is small relative to the
work involved in finding what encoding to use and doing error
checking, we \textit{compile} the encoding once for the attribute,
then evaluate the compiled form when encoding any value. All error
checking etc. is done during the compilation.

The general outline of the \texttt{encode} procedure is as follows.

@o acdc.tcl @{
proc acdc_encode { attr value } {
    @<get access to state information@>
    @<get the attribute id@>
    @<if there is no compiled form for the encoding, make one@>
    @<evaluate the compiled encoding@>
}
@}

As ususal we need to access the \texttt{acdc\_priv} variable and to
map \texttt{attr} to the attribute's id.  Next we see if a compiled
form of the encoding exists.  The compiled encoding is accessed with a
\textit{handle}.  This is a unique string that is passed to the
\texttt{d\_attr} command along with a value to encode the value.  The
handle is stored in the \texttt{acdc\_priv} array.  If no handle
exists, we compile the encoding using the private procedure
\texttt{acdcCompile} described in the next section.

@d if there is no compiled form for the encoding, make one @{
if ![info exists acdc_priv($id,handle)] {
    acdcCompile $attr ;
}
@}

Once we have the encoding handle, we can call \texttt{d\_attr encode}
passing in the value to be encoded.

@d evaluate the compiled encoding @{
d_attr encode $acdc_priv($id,handle) $value ;
@}

\subsubsection{The \texttt{acdcCompile} command} 
\label{sec:acdcCompile}
This procedure is the one that compiles encoding methods and saves the
handle used to access the compiled form.  The only argument to
\texttt{acdcCompile} is the index of the attribute being compiled,
\texttt{attr}.

The overall structure of the procedure is described below.  The body
of the procedure is a large switch statement on the encoding
type. Each branch compiles the appropriate encoding method and sets a
variable called \texttt{handle} which contains the handle for the
compiled encoding.

@o acdc.tcl @{
proc acdcCompile { attr } {
    @<get access to state information@>
    @<get the attribute id@>
    set type	$acdc_priv($id,type)
    set method	$acdc_priv($id,method)
    switch -exact -- $method {
	@<compile the ``ignore'' encoding@>
	@<compile the ``copy'' encoding@>
	@<compile the ``nm-abs'' encoding@>
	@<compile the ``nm-sqr'' encoding@>
	@<compile the ``rectan'' encoding@>
	@<compile the ``0-up'' encoding@>
	@<compile the ``1-up'' encoding@>
	@<compile the ``zero-one'' encoding@>
	@<compile the ``minus one-plus one'' encoding@>
	@<compile the ``1-of-n'' encoding@>
	@<compile the ``therm'' encoding@>
    }
    set acdc_priv($id,handle) $handle ;
}
@}

The branch of the compilation switch that does the \texttt{ignore}
encoding creates an attribute handle with the proper type, and
requires no additional arguments.

@d compile the ``ignore'' encoding @{
ignore  {
    set handle [d_attr create $type ignore] ;
}
@}

The next encoding type is \texttt{copy}. This type of attribute also
requires no additional arguments.

@d compile the ``copy'' encoding @{
copy  {
    set handle [d_attr create $type copy] ;
}
@}

There are two types of normalization encodings: \texttt{nm-sqr}, that
normalizes to zero mean and unit variance; and \texttt{nm-abs}, that
normalizes to zero median and unit average absolute deviation from the
median. Both encodings can be implemented using the same basic
encoding \texttt{normalized}. The only difference is the values used
to do the normalization: mean or median, variance or absolute
deviation.  They can even be used where the \texttt{center} option is
given. The \texttt{center} must simply be subtracted from the
midpoint.

The \texttt{nm-abs} branch of the compilation switch checks that the
required arguments have been specified and if the \texttt{center}
option has been given. If \texttt{center} has been specified, it
subtracts it from the \texttt{median} value. It then sets
\texttt{handle} to the handle returned by \texttt{d\_attr}.

@d compile the ``nm-abs'' encoding @{
nm-abs  {
    set mu	$acdc_priv($id,median) ;
    set sigma	$acdc_priv($id,absDev)
    set center  $acdc_priv($id,center) ;

    if { [string compare $mu {}] == 0 } {
	error "no median given for \"$method\" encoded attribute \"$attr\"" ;
    }
    if { [string compare $sigma {}] == 0 } {
	error "no absDev given for \"$method\" encoded attribute \"$attr\"" ;
    }
    if { [string compare  $center {}] != 0 } {
	set mu [expr { $mu - $center }] ;
    }
    set handle [d_attr create $type normalized $mu $sigma]
}
@}

The same encoding procedures can be used for the \texttt{nm-sqr}
encoding: we just pass the mean and variance instead of median and
absolute deviation. The same format is also used for the branch of the
switch statement. 

@d compile the ``nm-sqr'' encoding @{
nm-sqr  {
    set mu	$acdc_priv($id,mean) ;
    set sigma	$acdc_priv($id,variance)
    set center  $acdc_priv($id,center) ;

    if { [string compare $mu {}] == 0 } {
	error "no mean given for \"$method\" encoded attribute \"$attr\"" ;
    }
    if { [string compare  $center {}] != 0 } {
	set mu [expr { $mu - $center }] ;
    }
    if { [string compare $sigma {}] == 0 } {
	error "no variance given for \"$method\" encoded attribute \"$attr\"" ;
    }
    set handle [d_attr create $type normalized $mu $sigma] ;
}
@}

When encoding \texttt{rectan} values, a single value is encoded as the
$\sin$ and $\cos$ of the value divided by $2\pi\mbox{unit}$, where
\texttt{unit} is the range of the values (e.g. for months of the year,
12). The \texttt{unit} value must be supplied.  Since this is a one-to-many
mapping, decoding is not possible.

@d compile the ``rectan'' encoding @{
rectan {
    if { [string compare $acdc_priv($id,unit) {}] == 0 } {
	error "no unit given for \"$method\" encoded attribute \"$attr\"" ;
    }
    set handle [d_attr create $type rectan $acdc_priv($id,unit)]
}
@}

The \texttt{0-up} branch of the compilation switch is
straight-forward.  The range is converted from its standard \delve\
form to a list using the \texttt{range list} command (also provided by
\delve).  Since \delve{} ranges can really be lists of ranges, we have
to iterate over them accumulating the full list.

@d compile the ``0-up'' encoding @{
0-up {
    set list	{} ;
    foreach range $acdc_priv($id,range) {
	if { [string compare [range type $range] "enumerated"] } {
	    eval lappend list [range list $range] ;
	} else {
	    lappend list $range ;
	}
    }
    set handle [d_attr create $type zero-up $list] ;
}
@}

Similarly for the \texttt{1-up} branch.
@d compile the ``1-up'' encoding @{
1-up {
    set list	{} ;
    foreach range $acdc_priv($id,range) {
	if { [string compare [range type $range] "enumerated"] } {
	    eval lappend list [range list $range] ;
	} else {
	    lappend list $range ;
	}
    }
    set handle [d_attr create $type one-up $list] ;
}
@}

The \texttt{0/1} branch of the switch calls \texttt{d\_attr} passing
in: the method \texttt{binary-passive}; the range of allowed values
converted from its standard \delve\ form to a list using the
\texttt{range list} command (also provided by \delve); and the passive
value (which is not optional).

@d compile the ``zero-one'' encoding @{
0/1 {
    set passive	$acdc_priv($id,passive) ;
    if { [string compare $passive {}] == 0 } {
	error "no passive value given for \"$method\" encoded attribute \"$attr\""
    }
    set list	{} ;
    foreach range $acdc_priv($id,range) {
	if { [string compare [range type $range] "enumerated"] } {
	    eval lappend list [range list $range] ;
	} else {
	    lappend list $range ;
	}
    }
    set handle [d_attr create $type binary-passive $list $passive] ;
}
@}

The \texttt{-1/+1} branch of the switch statement also sets the list
of decoded values to the range of the attribute. It requires no
further arguments.

@d compile the ``minus one-plus one'' encoding @{
-1/+1 {
    set list	{} ;
    foreach range $acdc_priv($id,range) {
	if { [string compare [range type $range] "enumerated"] } {
	    eval lappend list [range list $range] ;
	} else {
	    lappend list $range ;
	}
    }
    set handle [d_attr create $type binary-symmetric $list] ;
}
@}

The \texttt{1-of-n} branch of the switch statement also sets the list
of decoded values to the range of the attribute. If there is a passive
value, it also passes that argument.

@d compile the ``1-of-n'' encoding @{
1-of-n {
    set passive	$acdc_priv($id,passive) ;
    set list	{} ;
    foreach range $acdc_priv($id,range) {
	if { [string compare [range type $range] "enumerated"] } {
	    eval lappend list [range list $range] ;
	} else {
	    lappend list $range ;
	}
    }
    if { [string compare $passive {}] == 0 } {
	set handle [d_attr create $type one-of-n $list] ;
    } else {
	set handle [d_attr create $type one-of-n $list $passive] ;
    }
}
@}

For the \texttt{therm} encoding we must first create the full list of
values in the attribute's range, then calculate the scaling factor
based on the number of values in the range.

@d compile the ``therm'' encoding @{
therm {
    set list	{} ;
    foreach range $acdc_priv($id,range) {
	if { [string compare [range type $range] "enumerated"] } {
	    eval lappend list [range list $range] ;
	} else {
	    lappend list $range ;
	}
    }
    set n	[llength $list] ;
    if { $n < 2 } {
	error "can't use thermometer coding for attribute $id: it has too few values in its range" ;
    }
    switch $acdc_priv($id,scale) {
	"linear"	{ set scale [expr { 1.0/($n - 1) }] }
	"sqrt"		{ set scale [expr { 1.0/sqrt($n - 1) }] }
	"none"		{ set scale 1.0 }
	default		{ error "unknown scaling method for thermometer coding of attribute \"$id\"" }
    }
    set handle [d_attr create $type thermometer $list $scale] ;
}
@}

\section{The \texttt{acdc\_decode} command}
\subsection{Description}
\usage{acdc\_decode}{attr value}
Returns the decoded representation of the encoded \texttt{value} using
the encoding method and options specified for the attribute
\texttt{attr}.

\subsection{Implementation}
The procedure takes the same arguments as \texttt{acdc\_encode},
\texttt{attr}, the index of the attribute, and \texttt{value} the
encoded value to decode; and has identical structure as well:

@o acdc.tcl @{
proc acdc_decode { attr value } {
    @<get access to state information@>
    @<get the attribute id@>
    @<if there is no compiled form for the encoding, make one@>
    @<evaluate the compiled decoding@>
}
@}

The only difference in the procedure is that it calls \texttt{d\_attr
decode} instead of \texttt{d\_attr encode}:

@d evaluate the compiled decoding @{
d_attr decode $acdc_priv($id,handle) $value ;
@}

\section{The \texttt{acdc\_encodeFile} command}
\subsection{Description}
\usage{acdc\_encodeFile}{attrList fromFileId toFileId} 
The \texttt{acdc\_encodeFile} command is used to encode an entire file
of raw values for a set of attributes.  It can be significantly faster
than calling \texttt{acdc\_encode} for all values in the file.

The procedure takes a list of attributes, \texttt{attrList}, and two
file id's.  It reads lines from the first file id,
\texttt{fromFileId}, until the end of file is reached.  Each line read
from the file id should contain the raw attribute values to be encoded
(one value for each attribute in \texttt{attrList}).  The procedure
encodes the values and writes them to \texttt{toFileId}.

\subsection{Implementation}
The overall outline of the procedure is as follows:
@o acdc.tcl @{
proc acdc_encodeFile { attrList fromFileId toFileId } {
    @<get access to state information@>
    @<save the encoding handles in a private array@>
    @<iterate over all lines, encoding the raw values@>
}
@}

The first thing we do is to get the encoding handles for each
attribute and save them into an array for quick access.  We also find
out how many attributes there are, because we will need it for every
line in the input file.

@d save the encoding handles in a private array @{
set numAttrs 0 ;
foreach attr $attrList {
    @<get the attribute id@>
    @<if there is no compiled form for the encoding, make one@>
    set handle($numAttrs) $acdc_priv($id,handle) ;
    incr numAttrs ;
}
@}

We now read lines from \texttt{fromFileId} until end of file.  For
each line we check that we have the right number of values, then we
iterate over them, encoding properly, and print the result to the
output file id.

@d iterate over all lines, encoding the raw values @{
set lineNum 0 ;
while { [gets $fromFileId line] >= 0 } {
    if { [llength $line] != $numAttrs } {
	error "wrong \# of values at line $lineNum" ;
    }
    set result 	{} ;
    set idx	0 ;
    foreach value $line {
	append result " [d_attr encode $handle($idx) $value]" ;
	incr idx ;
    }
    puts $toFileId $result ;
    incr lineNum ;
}
@}

\section{The \texttt{acdc\_decodeFile} command}
\subsection{Description}
\usage{acdc\_decodeFile}{attrList fromFileId toFileId}
The \texttt{acdc\_decodeFile} command is used to encode an entire file
of encoded values for a set of attributes.  It can be significantly
faster than calling \texttt{acdc\_encode} for all values in the file.

The procedure takes a list of attributes, \texttt{attrList}, and two
file id's.  It reads lines from the first file id,
\texttt{fromFileId}, until the end of file is reached.  Each line read
from the file id should contain the encoded values for the attributes
to be decoded.  The procedure decodes the values and writes them to
\texttt{toFileId}.

\subsection{Implementation}
The overall outline of the procedure is as follows:
@o acdc.tcl @{
proc acdc_decodeFile { attrList fromFileId toFileId } {
    @<get access to state information@>
    @<save the encoding handles in a private array@>
    @<calculate the number of elements in each encoded representation@>
    @<iterate over all lines, decoding the encoded values@>
}
@}

We get the state information and save the encodings into an array just
as we did for \texttt{acdc\_encodeFile}.

Next, we have to calculate how many values from each input line are
going to be used for each attribute: different encoding schemes may
use different numbers of values.  We store the ranges of values for
each attribute in the arrays \texttt{first} and \texttt{last}.

@d calculate the number of elements in each encoded representation @{
set start 0 ;
for { set idx 0 } { $idx < $numAttrs } { incr idx } {
    set num		[llength [acdc_names [lindex $attrList $idx]]] ;
    set first($idx)	$start ;
    set last($idx)	[expr { $first($idx) + $num - 1 }] ;

    incr start $num ;
}
set numEncodings $start ;
@}

We now read lines from \texttt{fromFileId} until end of file.  For
each line we check that we have the right number of values, then we
iterate over them, encoding properly, and print the result to the
output file id.

@d iterate over all lines, decoding the encoded values @{
set lineNum 0 ;
while { [gets $fromFileId line] >= 0 } {
    set line [eval list $line] ;
    if { [llength $line] != $numEncodings } {
	error "wrong \# of values at line $lineNum" ;
    }
    set result 	{} ;
    for { set idx 0 } { $idx < $numAttrs } { incr idx } {
	set value	[lrange $line $first($idx) $last($idx)] ;
	if { [string compare $value "?"] == 0 } {
	    lappend result $value ;
	} else {
	    lappend result [d_attr decode $handle($idx) $value] ;
	}
    }
    puts $toFileId $result ;
    incr lineNum ;
}
@}

\section{acdc\_list}
\subsection{Description}
\usage{acdc\_list}{?attr ...?}
This utility returns, for each attribute passed in as an argument, a
list with three elements. The first element is the \texttt{type} of
the attribute. The second is the attribute's \texttt{id}, and the
final is a list of all aliases for the attribute. If no \texttt{attr}
option is given to the command, it returns a list containing the same
information for all encoded attributes.

\subsection{Implementation}
The overall structure of the procedure is:

@o acdc.tcl @{
proc acdc_list { args } {
    @<get access to state information@>
    @<if args is empty, set it to a list of all atributes@>
    @<initialize the result@>
    foreach attr $args {
	@<get the attribute id@>
	@<build a list of aliases@>
	@<append attribute information to the result@>
    }
    @<return the result@>
}
@}

If no arguments were given to the procedure, then we have to generate
a list of all encoded attributes. To do this we iterate over all the
\texttt{acdc\_id} keys in the state information, accumulating the
attribute id's. We then remove duplicates.  We have to be careful that
we don't try to iterate over the array if it doesn't exist yet.

@d if args is empty, set it to a list of all atributes @{
if { [lempty $args] && [info exists acdc_priv] } {
    foreach key [array names acdc_priv] {
	if [regexp "^acdc_id," $key] {
	    lappend args $acdc_priv($key) ;
	}
    }
    set args [lrmdups $args] ;
}
@}

We have to make sure that we always initialize the result. That way,
even if there are no attributes, we return a valid result.

@d initialize the result @{
set result {} ;
@}

When iterating over the attributes, we first make map the attribute to
an id. If there is no mapping we want to generate an error.  Next, we
find all aliases that map to the id. We don't add the alias if it is
the same string as the id.

@d build a list of aliases @{
set aliases {}
foreach key [array names acdc_priv] {
    if { [regexp "^acdc_id," $key] 
	 && [string compare $acdc_priv($key) $id] == 0
	 && [regexp {acdc_id,(.*)} $key all alias]
	 && [string compare $alias $id] != 0 } {
	lappend aliases $alias ;
    }
}
@}

Next we append all the required information to the result.

@d append attribute information to the result @{
lappend result [list $acdc_priv($id,type) $id $aliases] ;
@}

The final line of the procedure returns the result.

@d return the result @{
return $result ;
@}

\section{acdc\_names}
\subsection{Description}
\usage{acdc\_names}{attr}
The encoded representation of an attribute may be represented using
more than one value (for example when using the \texttt{1-of-n}
encoding). Sometimes it is useful to know how many values the
representation will use, or to have symbolic names for these
values. This procedure returns a list containing a name for each value
in the encoded representation. The names are generated from the
attribute's id and a unique string describing each value in the
representation. If the encoding uses only one value, the list will
contain only the id of the attribute.

\subsection{Implementation}
The implementation of \texttt{acdc\_names} performs a switch on the
encoding type of the attribute. Most encodings generate are
represented using a single value. For these cases, the procedure
simply returns the id of the attribute. However, the encodings
\texttt{rectan}, \texttt{1-of-n}, and \texttt{therm} use multiple
values. As well, the \texttt{ignore} encoding uses \textit{no}
values. For these cases, names for each column must be generated.

The outline of the procedure is straight-forward. It takes a single
argument: \texttt{attr}, the attribute.

@o acdc.tcl @{
proc acdc_names { attr } {
    @<get access to state information@>
    @<get the attribute id@>
    @<generate the names for the encoding@>
    @<return a list of the names@>
}
@}

As usual, it starts with obtaining the state information, and
attribute id. It then procedes with the real work, generating the
names. This is performed in a switch statement. Each branch of the
switch covers an encoding scheme, and sets a variable called
\texttt{names} to contain the list of names. As mentioned above, most
simply return the name of the attribute; however some are more
complicated

@d generate the names for the encoding @{
switch -exact -- $acdc_priv($id,method) {
    "copy"	-
    "nm-abs"	-
    "nm-sqr"	-
    "0-up"	-
    "1-up"	-
    "0/1"	-
    "-1/+1"	{
	set names	[list $id] ;
    }
    @<generate names for the ``ignore'' encoding@>
    @<generate names for the ``rectan'' encoding@>
    @<generate names for the ``1-of-n'' encoding@>
    @<generate names for the ``therm''  encoding@>
}
@}

The list of names for the \texttt{ignore} encoding is an empty list:
@d generate names for the ``ignore'' encoding @{
"ignore" {
    set names	"" ;
}
@}

The \texttt{rectan} encoding uses two values, the $\sin$ and $\cos$ of
the angle:  
@d generate names for the ``rectan'' encoding @{
"rectan" {
    set names   [list "sin(k*$id)" "cos(k*$id)"] ;
}
@}

The final two encodings \texttt{1-of-n}, and \texttt{therm} are
grouped together. They simply iterate over the range of the attribute,
generating a name for each dimension. The range is converted from its
standard \delve\ form to a list using the \texttt{range list} command
(also provided by \delve).  The only tricky part of the operation is
remembering to skip over the passive value.

@d generate names for the ``1-of-n'' encoding @{
1-of-n {
    set names	"" ;
    foreach range $acdc_priv($id,range) {
	if { [string compare [range type $range] "enumerated"] } {
	    set list [range list $range] ;
	} else {
	    set list [list $range] ;
	}
	foreach value $list {
	    if { [string compare $value $acdc_priv($id,passive)] != 0 } {
		lappend names "$id:$value" ;
	    }
	}
    }
}
@}

@d generate names for the ``therm''  encoding @{
therm 	 {
    set names	"" ;
    foreach range $acdc_priv($id,range) {
	if { [string compare [range type $range] "enumerated"] } {
	    set list [range list $range] ;
	} else {
	    set list [list $range] ;
	}
	foreach value $list {
	    lappend names "$id:$value" ;
	}
    }
    set names	[lrange $names 1 end] ;
}
@}

Finally, all that remains is to return the list of names:
@d return a list of the names @{
return $names ;
@}


\section{acdc\_save}
\subsection{Description}
\usage{acdc\_save}{?attr ...?}
This utility procedure returns a string that can be evaluated to
restore the encoding state of the attributes given on the command
line. If no attributes are specified, a string is returned that can
restore all attributes in the encoding list.

\subsection{Implementation}
The outline of the procedure is straight-forward:

@o acdc.tcl @{
proc acdc_save args {
    @<get information about attributes using \texttt{acdc\_list}@>
    @<initialize the result@>
    foreach list $info {
	@<append attribute creation command@>
	@<append attribute method commands@>
	@<append attribute option commands@>
    }
    @<return the trimmed result@>
}
@}

The first step is to get the information about the required
attributes. To do this we simply call \texttt{acdc\_list} passing in
the same arguments we were given.

@d get information about attributes using \texttt{acdc\_list} @{
set info [eval acdc_list $args] ;
@}

We then initialize the result and iterate over all of these attributes
appending the commands used to create them.

@d append attribute creation command @{
set type	[lindex $list 0] ;
set id		[lindex $list 1] ;
set aliases	[lindex $list 2] ;
append result [list acdc_new $type $id $aliases] ;
append result "\n"
@}

Next we append the commands to set the default and active encoding
methods:

@d append attribute method commands @{
set method [acdc_method $id -default] ;
append result [list acdc_method $id -default -- $method] ;
append result "\n"

set method [acdc_method $id] ;
append result [list acdc_method $id -- $method] ;
append result "\n"
@}

Then we append the commands to set all of the encoding options:

@d append attribute option commands @{
foreach option [list absDev center mean median passive range variance unit] {
    set value [acdc_options $id -$option] ;
    append result [list acdc_options $id -$option $value] ;
    append result "\n"
}
@}

Finally, we return the result. We use \texttt{string trim} to get rid
of any preceding or trailing new lines.

@d return the trimmed result @{
return [string trim $result] ;
@}

\end{document}
